"""labels.py
============
The label set classes.
"""

import datetime
import json
from io import IOBase, BytesIO
from tempfile import NamedTemporaryFile
import pyarrow as pa
import numpy as np
import pandas as pd
from typing import Any, Optional, Union, Callable, List, Dict, Tuple, Set, IO, cast
from typing_extensions import Literal
from uuid import uuid4
from .util import LabelAttrs, is_valid_number, is_valid_float

HAS_PILLOW = False
try:
    from PIL import Image  # type: ignore

    HAS_PILLOW = True
except ImportError:
    HAS_PILLOW = False


from .util import (
    DEFAULT_SENSOR_ID,
    _is_one_gb_available,
    add_object_user_attrs,
    assert_valid_name,
    create_temp_directory,
    mark_temp_directory_complete,
    extract_illegal_characters,
    POLYGON_VERTICES_KEYS,
    KEYPOINT_KEYS,
    MAX_FRAMES_PER_BATCH,
    InferenceEntryDict,
    FrameEmbeddingDict,
    CropEmbeddingDict,
    LabelType,
    InferenceFrameSummary,
    LabelFrameSummary,
    BaseLabelEntryDict,
    BaseLabelAssetDict,
    ResizeMode,
    UpdateType,
)

CustomMetricsEntry = Union[float, List[List[Union[float, int]]]]
CropType = Literal["GT", "Inference"]


class BaseLabelSet:
    """A frame containing inferences from an experiment.

    Args:
        frame_id (str): A unique id for this frame.
    """

    # TODO: Is this the most pythonic pattern?
    frame_id: str
    crop_type: CropType
    label_data: List[BaseLabelEntryDict]
    label_assets: List[BaseLabelAssetDict]
    _label_ids_set: Set[str]
    crop_embeddings: List[CropEmbeddingDict]
    _seen_label_class_ids: Set[int]
    reuse_latest_embedding: bool
    crop_embedding_dim: Optional[int]
    update_type: UpdateType

    def __init__(
        self,
        *,
        frame_id: str,
        crop_type: CropType,
        reuse_latest_embedding: Optional[bool] = False,
        update_type: UpdateType = "ADD",
    ) -> None:
        if not isinstance(frame_id, str):
            raise Exception("frame ids must be strings")

        illegal_frame_id_characters = extract_illegal_characters(frame_id)
        if illegal_frame_id_characters:
            raise Exception(
                f"frame ids cannot contain the following character(s): {illegal_frame_id_characters}"
            )

        self.frame_id = frame_id
        self.crop_type = crop_type
        self.label_data = []
        self.label_assets = []
        self.crop_embeddings = []
        self._label_ids_set = set()
        self._seen_label_class_ids = set()
        self.reuse_latest_embedding = (
            False if reuse_latest_embedding is None else reuse_latest_embedding
        )
        self.crop_embedding_dim = None
        self.update_type = update_type

    def add_crop_embedding(
        self, *, label_id: str, embedding: List[float], model_id: str = ""
    ) -> None:
        """Add a per inference crop embedding

        Args:
            label_id (str): [description]
            embedding (List[float]): A vector of floats of at least length 2.
            model_id (str, optional): The model id used to generate these embeddings. Defaults to "".
        """
        if len(embedding) <= 1:
            raise Exception("Length of embeddings should be at least 2.")

        if not self.crop_embedding_dim:
            self.crop_embedding_dim = len(embedding)
        elif self.crop_embedding_dim != len(embedding):
            raise Exception(
                f"Length of embeddings must be the same, existing embeddings are of dimension {self.crop_embedding_dim} but new embedding has dimension {len(embedding)}"
            )

        # TODO: check label id present
        if label_id not in self._label_ids_set:
            raise Exception(
                f"Attempted to add embeddings for {label_id} which has not been added to the frame yet"
            )
        existing_labels = [
            label for label in self.label_data if label["uuid"] == label_id
        ]
        self.crop_embeddings.append(
            {
                "uuid": label_id,
                "embedding": embedding,
                "model_id": model_id,
                "date_generated": str(datetime.datetime.now()),
            }
        )
        for label in existing_labels:
            label["reuse_latest_embedding"] = False

    def _all_label_classes(self) -> List[str]:
        return list(
            set(
                [data["label"] for data in self.label_data if data["label"] != "__mask"]
            )
        )

    def _validate_label_inputs(self, *, label_id: str) -> None:
        if not isinstance(label_id, str):
            raise Exception("label ids must be strings")

        illegal_label_id_characters = extract_illegal_characters(label_id)
        if illegal_label_id_characters:
            raise Exception(
                f"frame ids cannot contain the following character(s): {illegal_label_id_characters}"
            )

    def to_dict(self) -> Dict[str, Any]:
        """Convert this label set into a dictionary representation.

        Returns:
            dict: dictified frame
        """
        row = {
            "task_id": self.frame_id,
            "label_data": self.label_data,
            "type": self.crop_type,
        }

        return row

    def _to_summary(self) -> LabelFrameSummary:
        """Converts this frame to a lightweight summary dict for internal cataloging

        Returns:
            dict: lightweight summaried frame
        """
        label_counts: Dict[LabelType, int] = {}
        for label in self.label_data:
            if not label_counts.get(label["label_type"]):
                label_counts[label["label_type"]] = 0
            label_counts[label["label_type"]] += 1

        return {
            "frame_id": self.frame_id,
            "label_counts": label_counts,
            "update_type": self.update_type,
        }

    def _validate_partial_args(
        self, label_id: str, classification: Optional[str] = None
    ) -> None:
        if not isinstance(label_id, str):
            raise Exception("label ids must be strings")
        if "/" in label_id:
            raise Exception("label ids cannot contain slashes (/)")

        if classification is not None and not isinstance(classification, str):
            raise Exception("classifications must be strings")

    def _append_partial_label(self, label_id: str, label_data: Dict[str, Any]) -> None:
        pruned_label_data = {k: v for k, v in label_data.items() if v is not None}
        pruned_label_data["change"] = "MODIFY"
        self.label_data.append(cast(BaseLabelEntryDict, pruned_label_data))
        self._label_ids_set.add(label_id)

    # Handle confidence
    def add_2d_line_segment(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        x1: Union[int, float],
        y1: Union[int, float],
        x2: Union[int, float],
        y2: Union[int, float],
        confidence: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a 2D line segment.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            x1 (int or float): The x-coord of the first vertex in pixels
            y1 (int or float): The x-coord of the first vertex in pixels
            x2 (int or float): The x-coord of the first vertex in pixels
            y2 (int or float): The x-coord of the first vertex in pixels
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_label_inputs(label_id=label_id)

        if not isinstance(classification, str):
            raise Exception("classifications must be strings")

        for coord in [x1, y1, x2, y2]:
            if not is_valid_number(coord):
                raise Exception("all line segment coordinates must be valid int/float")

        attrs = {"x1": x1, "y1": y1, "x2": x2, "y2": y2}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        if iscrowd is not None:
            attrs["iscrowd"] = iscrowd

        if links is not None:
            for k, v in links.items():
                if "link__" not in k:
                    k = "link__" + k
                attrs[k] = v

        add_object_user_attrs(attrs, user_attrs)

        self.label_data.append(
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "LINE_SEGMENT_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
                "change": "ADD",
            }
        )
        self._label_ids_set.add(label_id)

    def _update_2d_line_segment(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        classification: Optional[str] = None,
        x1: Optional[Union[int, float]] = None,
        y1: Optional[Union[int, float]] = None,
        x2: Optional[Union[int, float]] = None,
        y2: Optional[Union[int, float]] = None,
        confidence: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Partially update an existing 2D line segment. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            x1 (int or float, optional): The x-coord of the first vertex in pixels
            y1 (int or float, optional): The x-coord of the first vertex in pixels
            x2 (int or float, optional): The x-coord of the first vertex in pixels
            y2 (int or float, optional): The x-coord of the first vertex in pixels
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_partial_args(label_id, classification)

        for coord in [x1, y1, x2, y2]:
            if coord is not None and not is_valid_number(coord):
                raise Exception("all line segment coordinates must be valid int/float")

        coord_attrs = {"x1": x1, "y1": y1, "x2": x2, "y2": y2}
        attrs = {k: v for k, v in coord_attrs.items() if v is not None}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        if iscrowd is not None:
            attrs["iscrowd"] = iscrowd

        if links is not None:
            for k, v in links.items():
                if "link__" not in k:
                    k = "link__" + k
                attrs[k] = v

        add_object_user_attrs(attrs, user_attrs)
        self._append_partial_label(
            label_id,
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "LINE_SEGMENT_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
            },
        )

    def add_2d_bbox(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        top: Union[int, float],
        left: Union[int, float],
        width: Union[int, float],
        height: Union[int, float],
        confidence: Optional[float] = None,
        area: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a 2D bounding box.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            top (int or float): The top of the box in pixels
            left (int or float): The left of the box in pixels
            width (int or float): The width of the box in pixels
            height (int or float): The height of the box in pixels
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            area (float, optional): The area of the image.
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_label_inputs(label_id=label_id)

        if not isinstance(classification, str):
            raise Exception("classifications must be strings")

        for dim in [top, left, width, height]:
            if not is_valid_number(dim):
                raise Exception("all bounding box dimensions must be valid int/float")

        attrs = {
            "top": top,
            "left": left,
            "width": width,
            "height": height,
        }

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        if iscrowd is not None:
            attrs["iscrowd"] = iscrowd
        # TODO: This is mostly legacy for vergesense
        if area is not None:
            if not is_valid_float(area):
                raise Exception("area must be floats")
            if area < 0.0:
                raise Exception("area must be positive")
            attrs["area"] = area

        if links is not None:
            for k, v in links.items():
                if "link__" not in k:
                    k = "link__" + k
                attrs[k] = v

        add_object_user_attrs(attrs, user_attrs)

        self.label_data.append(
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "BBOX_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
                "change": "ADD",
            }
        )
        self._label_ids_set.add(label_id)

    def _update_2d_bbox(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str],
        classification: Optional[str],
        top: Optional[Union[int, float]],
        left: Optional[Union[int, float]],
        width: Optional[Union[int, float]],
        height: Optional[Union[int, float]],
        confidence: Optional[float] = None,
        area: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Update an existing 2D bounding box. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            top (int or float, optional): The top of the box in pixels
            left (int or float, optional): The left of the box in pixels
            width (int or float, optional): The width of the box in pixels
            height (int or float, optional): The height of the box in pixels
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            area (float, optional): The area of the image.
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_partial_args(label_id, classification)

        for dim in [top, left, width, height]:
            if dim is not None and not is_valid_number(dim):
                raise Exception("bounding box dimensions must be valid int/float")

        dim_attrs = {
            "top": top,
            "left": left,
            "width": width,
            "height": height,
        }
        attrs = {k: v for k, v in dim_attrs.items() if v is not None}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        if iscrowd is not None:
            attrs["iscrowd"] = iscrowd
        # TODO: This is mostly legacy for vergesense
        if area is not None:
            if not is_valid_float(area):
                raise Exception("area must be floats")
            if area < 0.0:
                raise Exception("area must be positive")
            attrs["area"] = area

        if links is not None:
            for k, v in links.items():
                if "link__" not in k:
                    k = "link__" + k
                attrs[k] = v

        add_object_user_attrs(attrs, user_attrs)

        self._append_partial_label(
            label_id,
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "BBOX_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
            },
        )

    def add_text_token(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        index: int,
        token: str,
        classification: str,
        visible: bool,
        confidence: Optional[float] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for a text token.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            index (int): the index of this token in the text
            token (str): the text content of this token
            classification (str): the classification string
            visible (bool): is this a visible token in the text
            confidence (float): confidence of prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self._validate_label_inputs(label_id=label_id)

        if not isinstance(classification, str):
            raise Exception("classifications must be strings")

        attrs = {
            "index": index,
            "token": token,
            "visible": visible,
        }

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        add_object_user_attrs(attrs, user_attrs)

        self.label_data.append(
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "TEXT_TOKEN",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
                "change": "ADD",
            }
        )
        self._label_ids_set.add(label_id)

    def _update_text_token(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        index: Optional[int] = None,
        token: Optional[str] = None,
        classification: Optional[str] = None,
        visible: Optional[bool] = None,
        confidence: Optional[float] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Partially update an existing label for a text token. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            index (int, optional): the index of this token in the text
            token (str, optional): the text content of this token
            classification (str, optional): the classification string
            visible (bool, optional): is this a visible token in the text
            confidence (float, optional): confidence of prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self._validate_partial_args(label_id, classification)

        token_attrs = {
            "index": index,
            "token": token,
            "visible": visible,
        }
        attrs: LabelAttrs = {k: v for k, v in token_attrs.items() if v is not None}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        add_object_user_attrs(attrs, user_attrs)

        self._append_partial_label(
            label_id,
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "TEXT_TOKEN",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
            },
        )

    def add_3d_cuboid(
        self,
        *,
        label_id: str,
        classification: str,
        position: List[float],
        dimensions: List[float],
        rotation: List[float],
        confidence: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
        coord_frame_id: Optional[str] = None,
    ) -> None:
        """Add an inference for a 3D cuboid.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            position (list of float): the position of the center of the cuboid
            dimensions (list of float): the dimensions of the cuboid
            rotation (list of float): the local rotation of the cuboid, represented as an xyzw quaternion.
            confidence (float): confidence of prediction
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
            links (dict, optional): Links between labels. Defaults to None.
            coord_frame_id (str, optional): Coordinate frame id. Defaults to 'world'
        """
        if coord_frame_id is None:
            coord_frame_id = "world"

        self._validate_label_inputs(label_id=label_id)

        if not isinstance(classification, str):
            raise Exception("classifications must be strings")

        attrs: Dict[str, Any] = {
            "pos_x": position[0],
            "pos_y": position[1],
            "pos_z": position[2],
            "dim_x": dimensions[0],
            "dim_y": dimensions[1],
            "dim_z": dimensions[2],
            "rot_x": rotation[0],
            "rot_y": rotation[1],
            "rot_z": rotation[2],
            "rot_w": rotation[3],
        }

        for k, v in attrs.items():
            if not is_valid_float(v):
                raise Exception(k + " must be valid float")

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        if iscrowd is not None:
            attrs["iscrowd"] = iscrowd

        add_object_user_attrs(attrs, user_attrs)

        if links is not None:
            for k, v in links.items():
                if "link__" not in k:
                    k = "link__" + k
                attrs[k] = v

        self.label_data.append(
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "CUBOID_3D",
                "label": classification,
                "label_coordinate_frame": coord_frame_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
                "change": "ADD",
            }
        )
        self._label_ids_set.add(label_id)

    def _update_3d_cuboid(
        self,
        *,
        label_id: str,
        classification: Optional[str] = None,
        position: Optional[List[float]] = None,
        dimensions: Optional[List[float]] = None,
        rotation: Optional[List[float]] = None,
        confidence: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
        coord_frame_id: Optional[str] = None,
    ) -> None:
        """Partially update an existing inference for a 3D cuboid. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str, optional): the classification string
            position (list of float, optional): the position of the center of the cuboid
            dimensions (list of float, optional): the dimensions of the cuboid
            rotation (list of float, optional): the local rotation of the cuboid, represented as an xyzw quaternion.
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
            links (dict, optional): Links between labels. Defaults to None.
            coord_frame_id (str, optional): Coordinate frame id. Defaults to 'world'
        """
        self._validate_partial_args(label_id, classification)

        attrs: Dict[str, Any] = {}
        if position is not None:
            attrs = {
                **attrs,
                "pos_x": position[0],
                "pos_y": position[1],
                "pos_z": position[2],
            }
        if dimensions is not None:
            attrs = {
                **attrs,
                "dim_x": dimensions[0],
                "dim_y": dimensions[1],
                "dim_z": dimensions[2],
            }
        if rotation is not None:
            attrs = {
                **attrs,
                "rot_x": rotation[0],
                "rot_y": rotation[1],
                "rot_z": rotation[2],
                "rot_w": rotation[3],
            }
        for k, v in attrs.items():
            if not is_valid_float(v):
                raise Exception(k + " must be valid float")

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        if iscrowd is not None:
            attrs["iscrowd"] = iscrowd

        add_object_user_attrs(attrs, user_attrs)

        if links is not None:
            for k, v in links.items():
                if "link__" not in k:
                    k = "link__" + k
                attrs[k] = v

        self._append_partial_label(
            label_id,
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "CUBOID_3D",
                "label": classification,
                "label_coordinate_frame": coord_frame_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
            },
        )

    def add_2d_keypoints(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        keypoints: List[Dict[KEYPOINT_KEYS, Union[int, float, str]]],
        confidence: Optional[float] = None,
        top: Optional[Union[int, float]] = None,
        left: Optional[Union[int, float]] = None,
        width: Optional[Union[int, float]] = None,
        height: Optional[Union[int, float]] = None,
        polygons: Optional[
            List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]]
        ] = None,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add an inference for a 2D keypoints task.

        A keypoint is a dictionary of the form:
            'x': x-coordinate in pixels
            'y': y-coordinate in pixels
            'name': string name of the keypoint

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            keypoints (list of dicts): The keypoints of this detection
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            top (int or float, optional): The top of the bounding box in pixels. Defaults to None.
            left (int or float, optional): The left of the bounding box in pixels. Defaults to None.
            width (int or float, optional): The width of the bounding box in pixels. Defaults to None.
            height (int or float, optional): The height of the bounding box in pixels. Defaults to None.
            polygons (list of dicts, optional): The polygon geometry. Defaults to None.
            center (list of ints or floats, optional): The center point of the polygon instance. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_label_inputs(label_id=label_id)

        if not isinstance(classification, str):
            raise Exception("classifications must be strings")

        keypoint_names_set = set()
        for keypoint in keypoints:
            x = keypoint["x"]
            y = keypoint["y"]
            name = keypoint["name"]
            if not is_valid_number(x):
                raise Exception("keypoint x coordinate must be valid int/float")
            if not is_valid_number(y):
                raise Exception("keypoint y coordinate must be valid int/float")
            if not isinstance(name, str) or len(name) == 0:
                raise Exception("keypoint name must be non-zero length string")
            if name in keypoint_names_set:
                raise Exception("keypoint names must be unique within a label")
            keypoint_names_set.add(name)

        attrs: Dict[str, Any] = {
            "keypoints": keypoints,
        }

        if (
            top is not None
            and left is not None
            and width is not None
            and height is not None
        ):
            for dim in [top, left, width, height]:
                if not is_valid_number(dim):
                    raise Exception(
                        "all bounding box dimensions must be valid int/float"
                    )
            attrs["top"] = top
            attrs["left"] = left
            attrs["width"] = width
            attrs["height"] = height

        if polygons:
            for polygon in polygons:
                for vertex in polygon["vertices"]:
                    for num in vertex:
                        if not is_valid_number(num):
                            raise Exception(
                                "all polygon vertex coordinates must be valid int/float"
                            )
            attrs["polygons"] = polygons

        if center is not None:
            for num in center:
                if not is_valid_number(num):
                    raise Exception(
                        "polygon center coordinates must be valid int/float"
                    )
            attrs["center"] = center

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        add_object_user_attrs(attrs, user_attrs)

        self.label_data.append(
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "KEYPOINTS_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
                "change": "ADD",
            }
        )
        self._label_ids_set.add(label_id)

    def _update_2d_keypoints(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        classification: Optional[str] = None,
        keypoints: Optional[List[Dict[KEYPOINT_KEYS, Union[int, float, str]]]] = None,
        confidence: Optional[float] = None,
        top: Optional[Union[int, float]] = None,
        left: Optional[Union[int, float]] = None,
        width: Optional[Union[int, float]] = None,
        height: Optional[Union[int, float]] = None,
        polygons: Optional[
            List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]]
        ] = None,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Partially update an inference for a 2D keypoints task. If the label_id does not already exist on the frame, it will be dropped.

        A keypoint is a dictionary of the form:
            'x': x-coordinate in pixels
            'y': y-coordinate in pixels
            'name': string name of the keypoint

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            top (int or float, optional): The top of the box in pixels
            left (int or float, optional): The left of the box in pixels
            width (int or float, optional): The width of the box in pixels
            height (int or float, optional): The height of the box in pixels
            keypoints (list of dicts, optional): The keypoints of this detection
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_partial_args(label_id, classification)

        attrs: Dict[str, Any] = {}
        keypoint_names_set = set()
        if keypoints is not None:
            for keypoint in keypoints:
                x = keypoint["x"]
                y = keypoint["y"]
                name = keypoint["name"]
                if not is_valid_number(x):
                    raise Exception("keypoint x coordinate must be valid int/float")
                if not is_valid_number(y):
                    raise Exception("keypoint y coordinate must be valid int/float")
                if not isinstance(name, str) or len(name) == 0:
                    raise Exception("keypoint name must be non-zero length string")
                if name in keypoint_names_set:
                    raise Exception("keypoint names must be unique within a label")
                keypoint_names_set.add(name)
            attrs["keypoints"] = keypoints

        for dim in [top, left, width, height]:
            if dim is not None and not is_valid_number(dim):
                raise Exception("all bounding box dimensions must be valid int/float")

        if top is not None:
            attrs["top"] = top
        if left is not None:
            attrs["left"] = left
        if width is not None:
            attrs["width"] = width
        if height is not None:
            attrs["height"] = height

        if polygons:
            for polygon in polygons:
                for vertex in polygon["vertices"]:
                    for num in vertex:
                        if not is_valid_number(num):
                            raise Exception(
                                "all polygon vertex coordinates must be valid int/float"
                            )
            attrs["polygons"] = polygons

        if center is not None:
            for num in center:
                if not is_valid_number(num):
                    raise Exception(
                        "polygon center coordinates must be valid int/float"
                    )
            attrs["center"] = center

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        add_object_user_attrs(attrs, user_attrs)

        self._append_partial_label(
            label_id,
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "KEYPOINTS_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
            },
        )

    def add_2d_polygon_list(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        polygons: List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]],
        confidence: Optional[float] = None,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add an inference for a 2D polygon list instance segmentation task.

        Polygons are dictionaries of the form:
            'vertices': List of (x, y) vertices (e.g. [[x1,y1], [x2,y2], ...])
                The polygon does not need to be closed with (x1, y1).
                As an example, a bounding box in polygon representation would look like:

                .. code-block::

                    {
                        'vertices': [
                            [left, top],
                            [left + width, top],
                            [left + width, top + height],
                            [left, top + height]
                        ]
                    }


        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            polygons (list of dicts): The polygon geometry
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            center (list of ints or floats, optional): The center point of the instance
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_label_inputs(label_id=label_id)

        if not isinstance(classification, str):
            raise Exception("classifications must be strings")

        for polygon in polygons:
            for vertex in polygon["vertices"]:
                for num in vertex:
                    if not is_valid_number(num):
                        raise Exception(
                            "all polygon vertex coordinates must be valid int/float"
                        )

        if center is not None:
            for num in center:
                if not is_valid_number(num):
                    raise Exception(
                        "polygon center coordinates must be valid int/float"
                    )

        attrs: Dict[str, Any] = {"polygons": polygons, "center": center}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        add_object_user_attrs(attrs, user_attrs)

        self.label_data.append(
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "POLYGON_LIST_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
                "change": "ADD",
            }
        )
        self._label_ids_set.add(label_id)

    def _update_2d_polygon_list(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        classification: Optional[str] = None,
        polygons: Optional[
            List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]]
        ] = None,
        confidence: Optional[float] = None,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Partially update an inference for a 2D polygon list instance segmentation task. If the label_id does not already exist on the frame, it will be dropped.

        Polygons are dictionaries of the form:
            'vertices': List of (x, y) vertices (e.g. [[x1,y1], [x2,y2], ...])
                The polygon does not need to be closed with (x1, y1).
                As an example, a bounding box in polygon representation would look like:

                .. code-block::

                    {
                        'vertices': [
                            [left, top],
                            [left + width, top],
                            [left + width, top + height],
                            [left, top + height]
                        ]
                    }


        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            polygons (list of dicts, optional): The polygon geometry
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            center (list of ints or floats, optional): The center point of the instance
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_partial_args(label_id, classification)

        if polygons is not None:
            for polygon in polygons:
                for vertex in polygon["vertices"]:
                    for num in vertex:
                        if not is_valid_number(num):
                            raise Exception(
                                "all polygon vertex coordinates must be valid int/float"
                            )

        if center is not None:
            for num in center:
                if not is_valid_number(num):
                    raise Exception(
                        "polygon center coordinates must be valid int/float"
                    )

        polygon_attrs: Dict[str, Any] = {"polygons": polygons, "center": center}
        attrs = {k: v for k, v in polygon_attrs.items() if v is not None}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        add_object_user_attrs(attrs, user_attrs)

        self._append_partial_label(
            label_id,
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "POLYGON_LIST_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
            },
        )

    # TODO: Do we want to expect column major or row major?
    def add_2d_semseg(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """
        Add an inference for 2D semseg. These should provide either a mask_url or a mask_data.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """

        self._validate_label_inputs(label_id=label_id)

        if mask_data is None and mask_url is None:
            raise Exception("Please provide either a mask_data or a mask_url")

        if mask_url is not None:
            self.label_data.append(
                {
                    "uuid": label_id,
                    "linked_labels": [],
                    "label": "__mask",
                    "label_coordinate_frame": sensor_id,
                    "label_type": "SEMANTIC_LABEL_URL_2D",
                    "attributes": {"url": mask_url, "resize_mode": resize_mode},
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                    "change": "ADD",
                }
            )
            self._label_ids_set.add(label_id)

        elif mask_data is not None:
            if not HAS_PILLOW:
                raise Exception(
                    "To use mask_data, please install with additional image dependencies using `pip install aquariumlearning[img]`"
                )

            if not isinstance(mask_data, (np.ndarray, np.generic)):
                raise Exception("mask_data must be a numpy ndarray")

            if not np.issubdtype(mask_data.dtype, np.integer):
                raise Exception("mask_data must have an integer dtype")

            if len(mask_data.shape) != 2:
                raise Exception(
                    "mask_data must be a 2d ndarray of shape [height, width]"
                )

            # TODO: Validate
            # TODO: Is png encoding too slow? Would something like pycocotools RLE be better?
            unique_label_ids = np.unique(mask_data)
            for v in unique_label_ids:
                self._seen_label_class_ids.add(cast(int, v))

            # Compress to a png for later usage, and immediate compression
            png_img = Image.fromarray(mask_data.astype("uint8"))
            png_io = BytesIO()
            png_img.save(
                png_io, format="PNG", compress_level=6
            )  # 0-9, defualt is 6, we favor slightly faster
            png_io.seek(0)
            png_bytes = png_io.read()

            asset_uuid = str(uuid4())
            self.label_assets.append(
                {"uuid": asset_uuid, "asset_type": "SEMSEG_MASK", "value": png_bytes}
            )

            self.label_data.append(
                {
                    "uuid": label_id,
                    "linked_labels": [],
                    "label": "__mask",
                    "label_coordinate_frame": sensor_id,
                    "label_type": "SEMANTIC_LABEL_ASSET_2D",
                    "attributes": {
                        "asset_uuid": asset_uuid,
                        "resize_mode": resize_mode,
                    },
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                    "change": "ADD",
                }
            )
            self._label_ids_set.add(label_id)

    # TODO: Do we want to expect column major or row major?
    def _update_2d_semseg(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """
        Update an inference for 2D semseg. These should provide either a mask_url or a mask_data.
        If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """

        self._validate_partial_args(label_id)

        if mask_url is not None:
            self._append_partial_label(
                label_id,
                {
                    "uuid": label_id,
                    "linked_labels": [],
                    "label": "__mask",
                    "label_coordinate_frame": sensor_id,
                    "label_type": "SEMANTIC_LABEL_URL_2D",
                    "attributes": {"url": mask_url, "resize_mode": resize_mode},
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                },
            )
        elif mask_data is not None:
            if not HAS_PILLOW:
                raise Exception(
                    "To use mask_data, please install with additional image dependencies using `pip install aquariumlearning[img]`"
                )

            if not isinstance(mask_data, (np.ndarray, np.generic)):
                raise Exception("mask_data must be a numpy ndarray")

            if not np.issubdtype(mask_data.dtype, np.integer):
                raise Exception("mask_data must have an integer dtype")

            if len(mask_data.shape) != 2:
                raise Exception(
                    "mask_data must be a 2d ndarray of shape [height, width]"
                )

            # TODO: Validate
            # TODO: Is png encoding too slow? Would something like pycocotools RLE be better?
            unique_label_ids = np.unique(mask_data)
            for v in unique_label_ids:
                self._seen_label_class_ids.add(cast(int, v))

            # Compress to a png for later usage, and immediate compression
            png_img = Image.fromarray(mask_data.astype("uint8"))
            png_io = BytesIO()
            png_img.save(
                png_io, format="PNG", compress_level=6
            )  # 0-9, defualt is 6, we favor slightly faster
            png_io.seek(0)
            png_bytes = png_io.read()

            asset_uuid = str(uuid4())
            self.label_assets.append(
                {"uuid": asset_uuid, "asset_type": "SEMSEG_MASK", "value": png_bytes}
            )

            self._append_partial_label(
                label_id,
                {
                    "uuid": label_id,
                    "linked_labels": [],
                    "label": "__mask",
                    "label_coordinate_frame": sensor_id,
                    "label_type": "SEMANTIC_LABEL_ASSET_2D",
                    "attributes": {
                        "asset_uuid": asset_uuid,
                        "resize_mode": resize_mode,
                    },
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                },
            )
        else:
            self._append_partial_label(
                label_id,
                {
                    "uuid": label_id,
                    "linked_labels": [],
                    "label": "__mask",
                    "label_coordinate_frame": sensor_id,
                    "label_type": "SEMANTIC_LABEL_URL_2D",
                    "attributes": {"resize_mode": resize_mode},
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                },
            )

    # TODO: Do we want to expect column major or row major?
    def add_2d_instance_seg(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        instance_mapping: List[Dict[str, Any]],
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """
        Add an label for 2D instance segmentation. These should provide either a mask_url or a mask_data, and a instance mapping.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            instance_mapping (List[Dict]): a list of instances present in the mask,
                each is a numeric `id`, string `classification`, and optional dict of additional attributes.
                As an example of one instance:
                .. code-block::
                        {
                            'id': 1,
                            'classification': "Person",
                            'attributes': {
                                'is_standing': false,
                            }
                        }
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """

        self._validate_label_inputs(label_id=label_id)

        if mask_data is None and mask_url is None:
            raise Exception("Please provide either a mask_data or a mask_url")

        for instance_info in instance_mapping:
            attrs = instance_info.get("attributes")
            if attrs is None:
                attrs = {}
            attrs["id"] = instance_info.get("id")
            attrs["resize_mode"] = resize_mode
            self.label_data.append(
                {
                    "uuid": label_id + "_" + str(instance_info.get("id")),
                    "linked_labels": [],
                    "label": instance_info.get("classification", ""),
                    "label_coordinate_frame": sensor_id,
                    "label_type": "INSTANCE_LABEL_2D",
                    "attributes": attrs,
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                    "change": "ADD",
                }
            )

        if mask_url is not None:
            self.label_data.append(
                {
                    "uuid": label_id,
                    "linked_labels": [],
                    "label": "__mask",
                    "label_coordinate_frame": sensor_id,
                    "label_type": "INSTANCE_LABEL_URL_2D",
                    "attributes": {"url": mask_url, "resize_mode": resize_mode},
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                    "change": "ADD",
                }
            )
            self._label_ids_set.add(label_id)

        elif mask_data is not None:
            if not HAS_PILLOW:
                raise Exception(
                    "To use mask_data, please install with additional image dependencies using `pip install aquariumlearning[img]`"
                )

            if not isinstance(mask_data, (np.ndarray, np.generic)):
                raise Exception("mask_data must be a numpy ndarray")

            if not np.issubdtype(mask_data.dtype, np.integer):
                raise Exception("mask_data must have an integer dtype")

            if len(mask_data.shape) != 2:
                raise Exception(
                    "mask_data must be a 2d ndarray of shape [height, width]"
                )

            # TODO: Validate
            # TODO: Is png encoding too slow? Would something like pycocotools RLE be better?
            unique_label_ids = np.unique(mask_data)
            for v in unique_label_ids:
                self._seen_label_class_ids.add(cast(int, v))

            # Compress to a png for later usage, and immediate compression
            png_img = Image.fromarray(mask_data.astype("uint8"))
            png_io = BytesIO()
            png_img.save(
                png_io, format="PNG", compress_level=6
            )  # 0-9, defualt is 6, we favor slightly faster
            png_io.seek(0)
            png_bytes = png_io.read()

            asset_uuid = str(uuid4())
            self.label_assets.append(
                {
                    "uuid": asset_uuid,
                    "asset_type": "INSTANCE_SEG_MASK",
                    "value": png_bytes,
                }
            )

            self.label_data.append(
                {
                    "uuid": label_id,
                    "linked_labels": [],
                    "label": "__mask",
                    "label_coordinate_frame": sensor_id,
                    "label_type": "INSTANCE_LABEL_ASSET_2D",
                    "attributes": {
                        "asset_uuid": asset_uuid,
                        "resize_mode": resize_mode,
                    },
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                    "change": "ADD",
                }
            )
            self._label_ids_set.add(label_id)

    # TODO: Do we want to expect column major or row major?
    def _update_2d_instance_seg(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        instance_mapping: Optional[List[Dict[str, Any]]] = None,
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """
        Update a label for 2D instance segmentation. These should provide either a mask_url or a mask_data, and a instance mapping.
        If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            instance_mapping (List[Dict]): a list of instances present in the mask,
                each is a numeric `id`, string `classification`, and optional dict of additional attributes.
                As an example of one instance:
                .. code-block::
                        {
                            'id': 1,
                            'classification': "Person",
                            'attributes': {
                                'is_standing': false,
                            }
                        }
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """

        self._validate_partial_args(label_id)

        if instance_mapping is not None:
            for instance_info in instance_mapping:
                attrs = instance_info.get("attributes")
                if attrs is None:
                    attrs = {}
                attrs["id"] = instance_info.get("id")
                attrs["resize_mode"] = resize_mode
                attrs = {k: v for k, v in attrs.items() if v is not None}
                self._append_partial_label(
                    label_id,
                    {
                        "uuid": label_id + "_" + str(instance_info.get("id")),
                        "linked_labels": [],
                        "label": instance_info.get("classification", ""),
                        "label_coordinate_frame": sensor_id,
                        "label_type": "INSTANCE_LABEL_2D",
                        "attributes": attrs,
                        "reuse_latest_embedding": self.reuse_latest_embedding,
                    },
                )

        if mask_url is not None:
            attrs = {"url": mask_url}
            if resize_mode is not None:
                attrs["resize_mode"] = resize_mode
            self._append_partial_label(
                label_id,
                {
                    "uuid": label_id,
                    "linked_labels": [],
                    "label": "__mask",
                    "label_coordinate_frame": sensor_id,
                    "label_type": "INSTANCE_LABEL_URL_2D",
                    "attributes": attrs,
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                },
            )

        elif mask_data is not None:
            if not HAS_PILLOW:
                raise Exception(
                    "To use mask_data, please install with additional image dependencies using `pip install aquariumlearning[img]`"
                )

            if not isinstance(mask_data, (np.ndarray, np.generic)):
                raise Exception("mask_data must be a numpy ndarray")

            if not np.issubdtype(mask_data.dtype, np.integer):
                raise Exception("mask_data must have an integer dtype")

            if len(mask_data.shape) != 2:
                raise Exception(
                    "mask_data must be a 2d ndarray of shape [height, width]"
                )

            # TODO: Validate
            # TODO: Is png encoding too slow? Would something like pycocotools RLE be better?
            unique_label_ids = np.unique(mask_data)
            for v in unique_label_ids:
                self._seen_label_class_ids.add(cast(int, v))

            # Compress to a png for later usage, and immediate compression
            png_img = Image.fromarray(mask_data.astype("uint8"))
            png_io = BytesIO()
            png_img.save(
                png_io, format="PNG", compress_level=6
            )  # 0-9, defualt is 6, we favor slightly faster
            png_io.seek(0)
            png_bytes = png_io.read()

            asset_uuid = str(uuid4())
            # TODO: modify to "replace" mask somehow?
            self.label_assets.append(
                {
                    "uuid": asset_uuid,
                    "asset_type": "INSTANCE_SEG_MASK",
                    "value": png_bytes,
                }
            )

            attrs = {"asset_uuid": asset_uuid}
            if resize_mode is not None:
                attrs["resize_mode"] = resize_mode
            self._append_partial_label(
                label_id,
                {
                    "uuid": label_id,
                    "linked_labels": [],
                    "label": "__mask",
                    "label_coordinate_frame": sensor_id,
                    "label_type": "INSTANCE_LABEL_ASSET_2D",
                    "attributes": attrs,
                    "reuse_latest_embedding": self.reuse_latest_embedding,
                },
            )

    def add_2d_classification(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        confidence: Optional[float] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        secondary_labels: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add an inference for 2D classification.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_label_inputs(label_id=label_id)

        if not isinstance(classification, str):
            raise Exception("classifications must be strings")

        attrs = {}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        if secondary_labels is not None:
            for k, v in secondary_labels.items():
                attrs[k] = v

        add_object_user_attrs(attrs, user_attrs)

        self.label_data.append(
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "CLASSIFICATION_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
                "change": "ADD",
            }
        )
        self._label_ids_set.add(label_id)

    def _update_2d_classification(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        classification: Optional[str],
        confidence: Optional[float] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        secondary_labels: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Update an existing label for 2D classification. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_partial_args(label_id, classification)

        attrs = {}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        if secondary_labels is not None:
            for k, v in secondary_labels.items():
                attrs[k] = v

        add_object_user_attrs(attrs, user_attrs)

        self._append_partial_label(
            label_id,
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "CLASSIFICATION_2D",
                "label": classification,
                "label_coordinate_frame": sensor_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
            },
        )

    def add_3d_classification(
        self,
        *,
        label_id: str,
        classification: str,
        confidence: Optional[float] = None,
        coord_frame_id: Optional[str] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for 3D classification.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            coord_frame_id (optional, str): The coordinate frame id.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        if coord_frame_id is None:
            coord_frame_id = "world"

        self._validate_label_inputs(label_id=label_id)

        if not isinstance(classification, str):
            raise Exception("classifications must be strings")

        attrs = {}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        add_object_user_attrs(attrs, user_attrs)

        self.label_data.append(
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "CLASSIFICATION_3D",
                "label": classification,
                "label_coordinate_frame": coord_frame_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
                "change": "ADD",
            }
        )
        self._label_ids_set.add(label_id)

    def _update_3d_classification(
        self,
        *,
        label_id: str,
        classification: Optional[str] = None,
        confidence: Optional[float] = None,
        coord_frame_id: Optional[str] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Update an existing label for 3D classification. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str, optional): the classification string
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            coord_frame_id (str, optional): The coordinate frame id.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._validate_partial_args(label_id, classification)

        attrs = {}

        if confidence is not None:
            if not is_valid_float(confidence):
                raise Exception("confidence must be floats")
            if confidence < 0.0 or confidence > 1.0:
                raise Exception("confidence must be between 0.0 and 1.0")

            attrs["confidence"] = confidence

        add_object_user_attrs(attrs, user_attrs)

        self._append_partial_label(
            label_id,
            {
                "uuid": label_id,
                "linked_labels": [],
                "label_type": "CLASSIFICATION_3D",
                "label": classification,
                "label_coordinate_frame": coord_frame_id,
                "attributes": attrs,
                "reuse_latest_embedding": self.reuse_latest_embedding,
            },
        )


class InferenceLabelSet(BaseLabelSet):
    custom_metrics: Dict[str, CustomMetricsEntry]
    visualization_assets: Dict[str, str]
    user_metadata: Dict[str, Any]
    reuse_latest_embedding: bool

    def __init__(
        self, *, frame_id: str, reuse_latest_embedding: Optional[bool] = False
    ) -> None:
        super().__init__(
            frame_id=frame_id,
            crop_type="Inference",
            reuse_latest_embedding=reuse_latest_embedding,
        )
        self.custom_metrics = {}
        self.visualization_assets = {}
        self.user_metadata = {}
        self.reuse_latest_embedding = (
            False if reuse_latest_embedding is None else reuse_latest_embedding
        )

    def add_custom_metric(self, name: str, value: CustomMetricsEntry) -> None:
        """Add a custom metric for a given inference frame.

        Args:
            name (str): The name of the custom metric being added. Must match one of the custom_metrics already defined by the corresponding Project.
            value (Union[float, List[List[Union[int, float]]]]): The value of the custom metric (either a float or 2d list of floats/integers).
        """
        if not (
            is_valid_float(value)
            or (
                isinstance(value, list)
                and all(all(is_valid_number(y) for y in x) for x in value)
            )
        ):
            raise Exception(
                "Custom metrics values must be either a float, or a 2D list of floats/integers."
            )

        self.custom_metrics[name] = value

    def add_visualization_asset(self, image_id: str, image_url: str) -> None:
        self.visualization_assets[image_id] = image_url

    def add_user_metadata(
        self,
        key: str,
        val: Union[str, int, float, bool],
    ) -> None:
        """Add a user provided metadata field.

        The types of these metadata fields will be infered and they'll be made
        available in the app for querying and metrics.

        Args:
            key (str): The key for your metadata field
            val (Union[str, int, float, bool]): value
            val_type (Literal["str", "int", "float", "bool"], optional): type of val as string. Defaults to None.
        """
        assert_valid_name(key)
        # Validates that neither val or type is None
        self.user_metadata[key] = val

    def to_dict(self) -> Dict[str, Any]:
        row = super().to_dict()

        row["custom_metrics"] = self.custom_metrics
        row["inference_metadata"] = {
            "visualization_assets": self.visualization_assets,
        }
        row["inference_data"] = row["label_data"]
        row["reuse_latest_embedding"] = self.reuse_latest_embedding
        del row["label_data"]

        row["inference_metadata"]["user_metadata"] = self.user_metadata

        return row

    def _to_summary(self) -> InferenceFrameSummary:
        """Converts this frame to a lightweight summary dict for internal cataloging

        Returns:
            dict: lightweight summaried frame
        """
        result: Any = super()._to_summary()
        result["custom_metrics_names"] = list(self.custom_metrics.keys())
        return result


class GTLabelSet(BaseLabelSet):
    def __init__(
        self,
        *,
        frame_id: str,
        update_type: UpdateType,
        reuse_latest_embedding: Optional[bool] = False,
    ) -> None:
        super().__init__(
            frame_id=frame_id,
            crop_type="GT",
            update_type=update_type,
            reuse_latest_embedding=reuse_latest_embedding,
        )


class UpdateGTLabelSet(GTLabelSet):
    is_snapshot: bool

    def __init__(self, *, frame_id: str, is_snapshot: Optional[bool] = True) -> None:
        super().__init__(
            frame_id=frame_id, update_type="MODIFY", reuse_latest_embedding=True
        )
        self.is_snapshot = True

    # TODO: figure out how to dedupe this vs LabeledFrame?
    def add_label_text_token(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        index: int,
        token: str,
        classification: str,
        visible: bool,
        confidence: Optional[float] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for a text token.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            index (int): the index of this token in the text
            token (str): the text content of this token
            classification (str): the classification string
            visible (bool): is this a visible token in the text
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_text_token(
            sensor_id=sensor_id,
            label_id=label_id,
            index=index,
            token=token,
            classification=classification,
            visible=visible,
            confidence=confidence,
            user_attrs=user_attrs,
        )

    def update_label_text_token(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        index: Optional[int] = None,
        token: Optional[str] = None,
        classification: Optional[str] = None,
        visible: Optional[bool] = None,
        confidence: Optional[float] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Partially update an existing label for a text token. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            index (int, optional): the index of this token in the text
            token (str, optional): the text content of this token
            classification (str, optional): the classification string
            visible (bool, optional): is this a visible token in the text
            confidence (float, optional): confidence of prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self._update_text_token(
            label_id=label_id,
            sensor_id=sensor_id,
            index=index,
            token=token,
            classification=classification,
            visible=visible,
            confidence=confidence,
            user_attrs=user_attrs,
        )

    def add_label_3d_cuboid(
        self,
        *,
        label_id: str,
        classification: str,
        position: List[float],
        dimensions: List[float],
        rotation: List[float],
        confidence: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
        coord_frame_id: Optional[str] = None,
    ) -> None:
        """Add a label for a 3D cuboid.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            position (list of float): the position of the center of the cuboid
            dimensions (list of float): the dimensions of the cuboid
            rotation (list of float): the local rotation of the cuboid, represented as an xyzw quaternion.
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
            links (dict, optional): Links between labels. Defaults to None.
            coord_frame_id (str, optional): Coordinate frame id. Defaults to 'world'
        """
        self.add_3d_cuboid(
            label_id=label_id,
            classification=classification,
            position=position,
            dimensions=dimensions,
            rotation=rotation,
            confidence=confidence,
            iscrowd=iscrowd,
            user_attrs=user_attrs,
            links=links,
            coord_frame_id=coord_frame_id,
        )

    def update_label_3d_cuboid(
        self,
        *,
        label_id: str,
        classification: Optional[str] = None,
        position: Optional[List[float]] = None,
        dimensions: Optional[List[float]] = None,
        rotation: Optional[List[float]] = None,
        confidence: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
        coord_frame_id: Optional[str] = None,
    ) -> None:
        """Partially update an existing inference for a 3D cuboid. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str, optional): the classification string
            position (list of float, optional): the position of the center of the cuboid
            dimensions (list of float, optional): the dimensions of the cuboid
            rotation (list of float, optional): the local rotation of the cuboid, represented as an xyzw quaternion.
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
            links (dict, optional): Links between labels. Defaults to None.
            coord_frame_id (str, optional): Coordinate frame id. Defaults to 'world'
        """
        self._update_3d_cuboid(
            label_id=label_id,
            classification=classification,
            position=position,
            dimensions=dimensions,
            rotation=rotation,
            confidence=confidence,
            iscrowd=iscrowd,
            user_attrs=user_attrs,
            links=links,
            coord_frame_id=coord_frame_id,
        )

    def add_label_2d_bbox(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        top: Union[int, float],
        left: Union[int, float],
        width: Union[int, float],
        height: Union[int, float],
        confidence: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for a 2D bounding box.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            top (int or float): The top of the box in pixels
            left (int or float): The left of the box in pixels
            width (int or float): The width of the box in pixels
            height (int or float): The height of the box in pixels
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
            links (dict, optional): Links between labels. Defaults to None.
        """
        self.add_2d_bbox(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            top=top,
            left=left,
            width=width,
            height=height,
            confidence=confidence,
            area=None,
            iscrowd=iscrowd,
            user_attrs=user_attrs,
            links=links,
        )

    def update_label_2d_bbox(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        classification: Optional[str] = None,
        top: Optional[Union[int, float]] = None,
        left: Optional[Union[int, float]] = None,
        width: Optional[Union[int, float]] = None,
        height: Optional[Union[int, float]] = None,
        confidence: Optional[float] = None,
        area: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Update an existing 2D bounding box. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            top (int or float, optional): The top of the box in pixels
            left (int or float, optional): The left of the box in pixels
            width (int or float, optional): The width of the box in pixels
            height (int or float, optional): The height of the box in pixels
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            area (float, optional): The area of the image.
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self._update_2d_bbox(
            label_id=label_id,
            sensor_id=sensor_id,
            classification=classification,
            top=top,
            left=left,
            width=width,
            height=height,
            confidence=confidence,
            area=None,
            iscrowd=iscrowd,
            user_attrs=user_attrs,
            links=links,
        )

    def add_label_2d_line_segment(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        x1: Union[int, float],
        y1: Union[int, float],
        x2: Union[int, float],
        y2: Union[int, float],
        confidence: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for a 2D line segment.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            x1 (int or float): The x-coord of the first vertex in pixels
            y1 (int or float): The x-coord of the first vertex in pixels
            x2 (int or float): The x-coord of the first vertex in pixels
            y2 (int or float): The x-coord of the first vertex in pixels
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_2d_line_segment(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            x1=x1,
            y1=y1,
            x2=x2,
            y2=y2,
            confidence=confidence,
            iscrowd=iscrowd,
            user_attrs=user_attrs,
            links=links,
        )

    def update_label_2d_line_segment(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        classification: Optional[str] = None,
        x1: Optional[Union[int, float]] = None,
        y1: Optional[Union[int, float]] = None,
        x2: Optional[Union[int, float]] = None,
        y2: Optional[Union[int, float]] = None,
        confidence: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Partially update an existing 2D line segment. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            x1 (int or float, optional): The x-coord of the first vertex in pixels
            y1 (int or float, optional): The x-coord of the first vertex in pixels
            x2 (int or float, optional): The x-coord of the first vertex in pixels
            y2 (int or float, optional): The x-coord of the first vertex in pixels
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self._update_2d_line_segment(
            label_id=label_id,
            sensor_id=sensor_id,
            classification=classification,
            x1=x1,
            y1=y1,
            x2=x2,
            y2=y2,
            confidence=confidence,
            iscrowd=iscrowd,
            user_attrs=user_attrs,
            links=links,
        )

    def add_label_2d_keypoints(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        keypoints: List[Dict[KEYPOINT_KEYS, Union[int, float, str]]],
        confidence: Optional[float] = None,
        top: Optional[Union[int, float]] = None,
        left: Optional[Union[int, float]] = None,
        width: Optional[Union[int, float]] = None,
        height: Optional[Union[int, float]] = None,
        polygons: Optional[
            List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]]
        ] = None,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for a 2D keypoints task.

        A keypoint is a dictionary of the form:
            'x': x-coordinate in pixels
            'y': y-coordinate in pixels
            'name': string name of the keypoint

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            keypoints (list of dicts): The keypoints of this detection
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            top (int or float, optional): The top of the bounding box in pixels. Defaults to None.
            left (int or float, optional): The left of the bounding box in pixels. Defaults to None.
            width (int or float, optional): The width of the bounding box in pixels. Defaults to None.
            height (int or float, optional): The height of the bounding box in pixels. Defaults to None.
            polygons (list of dicts, optional): The polygon geometry. Defaults to None.
            center (list of ints or floats, optional): The center point of the polygon instance. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_2d_keypoints(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            keypoints=keypoints,
            confidence=confidence,
            top=top,
            left=left,
            width=width,
            height=height,
            polygons=polygons,
            center=center,
            user_attrs=user_attrs,
        )

    def update_label_2d_keypoints(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        classification: Optional[str] = None,
        keypoints: Optional[List[Dict[KEYPOINT_KEYS, Union[int, float, str]]]] = None,
        confidence: Optional[float] = None,
        top: Optional[Union[int, float]] = None,
        left: Optional[Union[int, float]] = None,
        width: Optional[Union[int, float]] = None,
        height: Optional[Union[int, float]] = None,
        polygons: Optional[
            List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]]
        ] = None,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Partially update an inference for a 2D keypoints task. If the label_id does not already exist on the frame, it will be dropped.

        A keypoint is a dictionary of the form:
            'x': x-coordinate in pixels
            'y': y-coordinate in pixels
            'name': string name of the keypoint

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            top (int or float, optional): The top of the box in pixels
            left (int or float, optional): The left of the box in pixels
            width (int or float, optional): The width of the box in pixels
            height (int or float, optional): The height of the box in pixels
            keypoints (list of dicts, optional): The keypoints of this detection
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self._update_2d_keypoints(
            label_id=label_id,
            sensor_id=sensor_id,
            classification=classification,
            keypoints=keypoints,
            confidence=confidence,
            top=top,
            left=left,
            width=width,
            height=height,
            polygons=polygons,
            center=center,
            user_attrs=user_attrs,
        )

    def add_label_2d_polygon_list(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        polygons: List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]],
        confidence: Optional[float] = None,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for a 2D polygon list instance segmentation task.

        Polygons are dictionaries of the form:
            'vertices': List of (x, y) vertices (e.g. [[x1,y1], [x2,y2], ...])
                The polygon does not need to be closed with (x1, y1).
                As an example, a bounding box in polygon representation would look like:

                .. code-block::

                    {
                        'vertices': [
                            [left, top],
                            [left + width, top],
                            [left + width, top + height],
                            [left, top + height]
                        ]
                    }

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            polygons (list of dicts): The polygon geometry
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            center (list of ints or floats, optional): The center point of the instance
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_2d_polygon_list(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            polygons=polygons,
            confidence=confidence,
            center=center,
            user_attrs=user_attrs,
        )

    def update_label_2d_polygon_list(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        classification: Optional[str] = None,
        polygons: Optional[
            List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]]
        ] = None,
        confidence: Optional[float] = None,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Partially update an label for a 2D polygon list instance segmentation task. If the label_id does not already exist on the frame, it will be dropped.

        Polygons are dictionaries of the form:
            'vertices': List of (x, y) vertices (e.g. [[x1,y1], [x2,y2], ...])
                The polygon does not need to be closed with (x1, y1).
                As an example, a bounding box in polygon representation would look like:

                .. code-block::

                    {
                        'vertices': [
                            [left, top],
                            [left + width, top],
                            [left + width, top + height],
                            [left, top + height]
                        ]
                    }


        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            polygons (list of dicts, optional): The polygon geometry
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            center (list of ints or floats, optional): The center point of the instance
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self._update_2d_polygon_list(
            label_id=label_id,
            sensor_id=sensor_id,
            classification=classification,
            polygons=polygons,
            confidence=confidence,
            center=center,
            user_attrs=user_attrs,
        )

    def add_label_2d_instance_seg(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        instance_mapping: List[Dict[str, Any]],
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """Add a label for 2D instance segmentation.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            instance_mapping (List[Dict]): a list of instances present in the mask,
                each is a numeric `id`, string `classification`, and optional dict of additional attributes.
                As an example of one instance:
                .. code-block::
                        {
                            'id': 1,
                            'classification': "Person",
                            'attributes': {
                                'is_standing': false,
                            }
                        }
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """
        self.add_2d_instance_seg(
            sensor_id=sensor_id,
            label_id=label_id,
            mask_url=mask_url,
            mask_data=mask_data,
            instance_mapping=instance_mapping,
            resize_mode=resize_mode,
        )

    def update_label_2d_instance_seg(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        instance_mapping: Optional[List[Dict[str, Any]]] = None,
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """
        Update a label for 2D instance segmentation. These should provide either a mask_url or a mask_data, and a instance mapping.
        If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            instance_mapping (List[Dict]): a list of instances present in the mask,
                each is a numeric `id`, string `classification`, and optional dict of additional attributes.
                As an example of one instance:
                .. code-block::
                        {
                            'id': 1,
                            'classification': "Person",
                            'attributes': {
                                'is_standing': false,
                            }
                        }
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """
        self._update_2d_instance_seg(
            label_id=label_id,
            sensor_id=sensor_id,
            mask_url=mask_url,
            mask_data=mask_data,
            instance_mapping=instance_mapping,
            resize_mode=resize_mode,
        )

    def add_label_2d_semseg(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """Add a label for 2D semseg.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """
        self.add_2d_semseg(
            sensor_id=sensor_id,
            label_id=label_id,
            mask_url=mask_url,
            mask_data=mask_data,
            resize_mode=resize_mode,
        )

    def update_label_2d_semseg(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """
        Update an label for 2D semseg. These should provide either a mask_url or a mask_data.
        If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """
        self._update_2d_semseg(
            label_id=label_id,
            sensor_id=sensor_id,
            mask_url=mask_url,
            mask_data=mask_data,
            resize_mode=resize_mode,
        )

    def add_label_2d_classification(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        confidence: Optional[float] = None,
        secondary_labels: Optional[Dict[str, Any]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for 2D classification.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            secondary_labels (dict, optional): dictionary of secondary labels
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_2d_classification(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            secondary_labels=secondary_labels,
            confidence=confidence,
            user_attrs=user_attrs,
        )

    def update_label_2d_classification(
        self,
        *,
        label_id: str,
        sensor_id: Optional[str] = None,
        classification: Optional[str],
        confidence: Optional[float] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        secondary_labels: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Update an existing label for 2D classification.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            sensor_id (str, optional): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            classification (str, optional): the classification string
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self._update_2d_classification(
            label_id=label_id,
            sensor_id=sensor_id,
            classification=classification,
            secondary_labels=secondary_labels,
            confidence=confidence,
            user_attrs=user_attrs,
        )

    def add_label_3d_classification(
        self,
        *,
        label_id: str,
        classification: str,
        confidence: Optional[float] = None,
        coord_frame_id: Optional[str] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for 3D classification.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            confidence (float, optional): (Unlabeled Datasets Only) The confidence between 0.0 and 1.0 of the proposed label.
            coord_frame_id (str, optional): The coordinate frame id.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self.add_3d_classification(
            label_id=label_id,
            classification=classification,
            confidence=confidence,
            coord_frame_id=coord_frame_id,
            user_attrs=user_attrs,
        )

    def update_label_3d_classification(
        self,
        *,
        label_id: str,
        classification: Optional[str] = None,
        confidence: Optional[float] = None,
        coord_frame_id: Optional[str] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Update an existing label for 3D classification. If the label_id does not already exist on the frame, it will be dropped.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str, optional): the classification string
            confidence (float, optional): The confidence between 0.0 and 1.0 of the prediction
            coord_frame_id (str, optional): The coordinate frame id.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self._update_3d_classification(
            label_id=label_id,
            classification=classification,
            confidence=confidence,
            coord_frame_id=coord_frame_id,
            user_attrs=user_attrs,
        )

    def to_dict(self) -> Dict[str, Any]:
        if self.is_snapshot:
            return super().to_dict()
        else:
            raise Exception("Not Implemented")


class UpdateInferenceLabelSet(InferenceLabelSet):
    is_snapshot: bool

    def __init__(self, *, frame_id: str, is_snapshot: Optional[bool] = True) -> None:
        super().__init__(frame_id=frame_id)
        self.is_snapshot = True

    def to_dict(self) -> Dict[str, Any]:
        if self.is_snapshot:
            return super().to_dict()
        else:
            raise Exception("Not Implemented")


class UnlabeledInferenceSet(GTLabelSet):
    def __init__(
        self,
        *,
        frame_id: str,
        update_type: UpdateType,
        reuse_latest_embedding: Optional[bool] = False,
    ) -> None:
        super().__init__(
            frame_id=frame_id,
            update_type=update_type,
            reuse_latest_embedding=reuse_latest_embedding,
        )

    def add_crop_embedding(
        self, *, label_id: str, embedding: List[float], model_id: str = ""
    ) -> None:
        raise Exception(
            "Custom embeddings are not supported for this iteration of UnlabeledInferenceSet"
        )


# class Frame:

# class LabeledFrame:


# class FrameEmbedding:

# class CropEmbedding:



"""work_queues.py
============
Functionality related to work queue management
"""

import requests
from collections.abc import Iterable
from .util import all_issue_states, raise_resp_exception_error, ElementType, IssueState
from typing import Any, Union, List, Dict, Optional, TYPE_CHECKING
from typing_extensions import TypedDict

if TYPE_CHECKING:
    from .client import Client


class WorkQueueElement:
    """Definition for work queue element.

    Args:
        element_id (str): The element id.
        frame_id (str): The frame id of the element.
        element_type (str): The element type of the issue element ("frame" or "crop").
        dataset (str): The base dataset an element is from. (Can be formatted as either "project_name.dataset_name" or just "dataset_name")
        batch_name (str): The batch within the queue that the element belongs to
        status (str): The work status of the element. Can be queued, assigned, or done
        assignee: (Optional[str]): The email of the person who is assigned to work on the element
        inference_set (Optional[str]): The inference set an element is from (if any). (Can be formatted as either "project_name.inference_set_name" or just "inference_set_name")
        label_metadata (Optional[Dict[str, Any]]): (*For read purposes, not element modification*). JSON object with confidence and IOU info if the element was created from a ground truth/inference comparison.
    """

    def __init__(
        self,
        element_id: str,
        frame_id: str,
        element_type: str,
        dataset: str,
        batch_name: str,
        status: str,
        assignee: Optional[str] = None,
        inference_set: Optional[str] = None,
        label_metadata: Optional[Dict[str, Any]] = None,
    ):
        if element_type != "crop" and element_type != "frame":
            raise Exception('element_type must be either "crop" or "frame"')

        self.element_id = element_id
        self.frame_id = frame_id
        self.element_type = element_type
        self.status = status
        self.assignee = assignee
        self.batch_name = batch_name
        self.dataset = dataset
        self.inference_set = inference_set
        self.label_metadata = label_metadata

    def to_dict(self) -> Dict[str, Any]:
        return {
            "element_id": self.element_id,
            "frame_id": self.frame_id,
            "element_type": self.element_type,
            "status": self.status,
            "assignee": self.assignee,
            "batch_name": self.batch_name,
            "dataset": self.dataset,
            "inference_set": self.inference_set,
            "label_metadata": self.label_metadata,
        }


class WorkQueueApiResp(TypedDict):
    id: str
    compare_dataset: Optional[str]
    dataset: Optional[str]
    name: str
    element_type: ElementType
    created_at: Optional[str]
    updated_at: Optional[str]
    state: Optional[IssueState]
    issue_id: Optional[str]
    element_count: int
    batch_statuses: Dict[str, str]
    batch_assignees: Dict[str, str]
    elements: Optional[List[Dict[str, Any]]]


class WorkQueue:
    """Definition for work queue.

    Args:
        name (str): The work queue name.
        dataset (Optional[str]): The dataset for this work queue.
        elements (List[WorkQueueElement]): The elements of the work queue.
        element_type (str): The element type of the work queue ("frame", "crop").
        element_count (int): The number of elements in the work queue.
        created_at (str): The time of work queue creation.
        updated_at (str): The time of last work queue update.
        work_queue_id (str): The work queue id.
        batch_statuses (Dict[str, str]): The statuses of all batches in this queue.
        batch_assignees (Dict[str, str]): The emails of users assigned to batches if they have been assigned
        inference_set (Optional[str], optional): The inference set for this work queue. Defaults to None.
    """

    def __init__(
        self,
        name: str,
        dataset: Optional[str],
        element_count: Optional[int],
        element_type: ElementType,
        work_queue_id: str,
        batch_statuses: Dict[str, str],
        batch_assignees: Dict[str, str],
        elements: Optional[List[WorkQueueElement]] = None,
        # TODO: This used to say datetime, but it looks like it should be a str?
        created_at: Optional[str] = None,
        updated_at: Optional[str] = None,
        inference_set: Optional[str] = None,
    ):

        self.name = name
        self.dataset = dataset
        self.elements = elements
        self.element_count = element_count or (len(elements) if elements else 0)
        self.element_type = element_type
        self.batch_statuses = batch_statuses
        self.batch_assignees = batch_assignees
        self.created_at = created_at
        self.updated_at = updated_at
        self.work_queue_id = work_queue_id
        self.inference_set = inference_set

    def __repr__(self) -> str:
        return "Work Queue {} ({})".format(self.work_queue_id, self.name)

    def __str__(self) -> str:
        return "Work Queue {} ({})".format(self.work_queue_id, self.name)

    def to_dict(self) -> Dict[str, Any]:
        wq_dict: Dict[str, Any] = {
            "name": self.name,
            "dataset": self.dataset,
            "element_count": self.element_count,
            "element_type": self.element_type,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "work_queue_id": self.work_queue_id,
            "inference_set": self.inference_set,
            "batch_statuses": self.batch_statuses,
            "batch_assignees": self.batch_assignees,
        }
        if self.elements:
            wq_dict["elements"] = [x.to_dict() for x in self.elements]

        return wq_dict


class WorkQueueManager:
    """A work queue manager for interacting with work queues within a given project.

    Args:
        client (Client): An Aquarium Learning Python Client object.
        project_id (str): The project id associated with this manager.
    """

    def __init__(self, client: "Client", project_id: str) -> None:
        self.client = client
        self.project_id = project_id

    @staticmethod
    def _work_queue_from_api_resp(api_resp: WorkQueueApiResp) -> WorkQueue:
        # TODO: Hack because internal data model for work queues is still dataset/compare dataset,
        # not dataset + inference set + other inference set.

        compare_dataset = api_resp.get("compare_dataset")
        raw_dataset = api_resp.get("dataset")
        dataset: Optional[str] = None
        inference_set: Optional[str] = None

        batch_statuses = api_resp.get("batch_statuses", {})
        batch_assignees = api_resp.get("batch_assignees", {})

        for batch_name in batch_statuses:
            if batch_statuses[batch_name] != "done" and batch_name in batch_assignees:
                batch_statuses[batch_name] = "assigned"

        if compare_dataset and raw_dataset:
            dataset = compare_dataset.split(".")[1]
            inference_set = raw_dataset.split(".")[1]
        elif raw_dataset:
            dataset = raw_dataset.split(".")[1]
            inference_set = None
        else:  # in the case of an issue with no elements
            dataset = None
            inference_set = None

        elements = []
        raw_els = api_resp.get("elements", []) or []
        for raw_el in raw_els:
            # TODO: Is the change to explicit accesses vs nullable gets correct?
            elements.append(
                WorkQueueElement(
                    element_id=raw_el["element_id"],
                    element_type=api_resp["element_type"],
                    frame_id=raw_el["frame_id"],
                    status=raw_el["status"],
                    batch_name=raw_el["batch_name"],
                    assignee=raw_el["assignee"],
                    dataset=raw_el["dataset"],
                    inference_set=raw_el.get("inference_set"),
                    label_metadata=raw_el.get("label_metadata"),
                )
            )

        # TODO: Is the change to explicit accesses vs nullable gets correct?
        return WorkQueue(
            name=api_resp["name"],
            element_type=api_resp["element_type"],
            created_at=api_resp.get("created_at"),
            updated_at=api_resp.get("updated_at"),
            work_queue_id=api_resp["id"],
            dataset=dataset,
            inference_set=inference_set,
            element_count=api_resp.get("element_count"),
            elements=elements or None,
            batch_statuses=batch_statuses,
            batch_assignees=batch_assignees,
        )

    def list_work_queues(self) -> List[WorkQueue]:
        """List work queues in the associated project.

        NOTE: this does NOT include any element data, just the element counts.
        (Use `get_work_queue` instead to see that info).

        Returns:
            List[WorkQueue]: List of all work queues data.
        """
        url = "/projects/{}/issues/work_queues/summaries".format(self.project_id)
        r = requests.get(
            self.client.api_endpoint + url, headers=self.client._get_creds_headers()
        )

        raise_resp_exception_error(r)
        return [self._work_queue_from_api_resp(x) for x in r.json()]

    def get_work_queue(
        self,
        work_queue_id: str,
    ) -> WorkQueue:
        """Get a specific work queue in the associated project.

        Args:
            work_queue_id (str): The work queue id.

        Returns:
            WorkQueue: The work queue data (including elements).
        """
        url = "/projects/{}/issues/work_queues/{}/download_elements".format(
            self.project_id, work_queue_id
        )

        r = requests.get(
            self.client.api_endpoint + url,
            headers=self.client._get_creds_headers(),
            params={
                "get_as_gcs_file": True,
            },
        )

        raise_resp_exception_error(r)
        signed_url = r.json()["signed_url"]
        r = requests.get(signed_url)
        return self._work_queue_from_api_resp(r.json())



"""inference.py
============
The inference and inference frame modules.
"""

import datetime
import json
import numbers
from io import IOBase
from tempfile import NamedTemporaryFile
import pyarrow as pa
import numpy as np
import pandas as pd
from typing import Any, Optional, Union, Callable, List, Dict, Tuple, Set, IO, cast
from typing_extensions import Protocol


from .util import (
    DUMMY_FRAME_EMBEDDING,
    DEFAULT_SENSOR_ID,
    ResizeMode,
    UpdateType,
    _is_one_gb_available,
    create_temp_directory,
    mark_temp_directory_complete,
    POLYGON_VERTICES_KEYS,
    KEYPOINT_KEYS,
    MAX_FRAMES_PER_BATCH,
    InferenceEntryDict,
    FrameEmbeddingDict,
    CropEmbeddingDict,
    LabelType,
    InferenceFrameSummary,
)

from .labels import InferenceLabelSet


class InferenceFrameWithEmbeddings(Protocol):
    embedding: FrameEmbeddingDict
    crop_embeddings: List[CropEmbeddingDict]


class InferencesFrame(InferenceLabelSet):
    embedding: Optional[FrameEmbeddingDict]
    frame_embedding_dim: int
    update_type: UpdateType

    def __init__(
        self,
        *,
        frame_id: str,
        reuse_latest_embedding: Optional[bool] = False,
        update_type: UpdateType = "ADD",
    ) -> None:
        super().__init__(
            frame_id=frame_id, reuse_latest_embedding=reuse_latest_embedding
        )
        self.embedding = None
        self.frame_embedding_dim = -1
        self.update_type = update_type

    def add_frame_embedding(
        self, *, embedding: List[float], model_id: str = ""
    ) -> None:
        """Add an embedding to this frame

        Args:
            embedding (List[float]): A vector of floats of at least length 2.
            model_id (str, optional): The model id used to generate these embeddings. Defaults to "".
        """
        if len(embedding) <= 1:
            raise Exception("Length of embeddings should be at least 2.")

        for embedding_el in embedding:
            if not isinstance(embedding_el, numbers.Number):
                raise Exception(
                    f"Unexpectedly encountered a {type(embedding[0])} element. Only flat arrays of numbers are supported for embeddings."
                )

        if not self.embedding:
            self.embedding = {
                "task_id": self.frame_id,
                "model_id": model_id,
                "date_generated": str(datetime.datetime.now()),
                "embedding": embedding,
            }
            self.frame_embedding_dim = len(embedding)

    def to_dict(self) -> Dict[str, Any]:
        row = super().to_dict()
        row["change"] = self.update_type
        return row

    def validate_and_backfill_crop_embeddings(self) -> bool:
        """Determines if there are frame embeddings but not Crop embeddings
        Backfills crop embeddings if Frame embeddings exist
        Returns True if backfill was necessary

        Returns:
            bool: returns if backfill on frame was necessary
        """

        # Handle case where Frame embedding is still None
        if not self.embedding:
            return False

        if len(self.crop_embeddings) > 0:
            return False

        return_flag = False

        for label in self.label_data:
            if label.get("label_type", "") in [
                "CLASSIFICATION_2D",
                "CLASSIFICATION_3D",
            ]:
                return False
            label_id = label["uuid"]

            self.add_crop_embedding(
                label_id=label_id, embedding=[0.0, 0.0, 0.0], model_id="crop_default"
            )
            return_flag = True
        return return_flag

    def add_embedding(
        self,
        *,
        embedding: List[float],
        crop_embeddings: Optional[List[CropEmbeddingDict]] = None,
        model_id: str = "",
    ) -> None:
        raise Exception("Deprecated")

    def add_inference_text_token(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        index: int,
        token: str,
        classification: str,
        visible: bool,
        confidence: float,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for a text token.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            index (int): the index of this token in the text
            token (str): the text content of this token
            classification (str): the classification string
            visible (bool): is this a visible token in the text
            confidence (float): confidence of prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_text_token(
            sensor_id=sensor_id,
            label_id=label_id,
            index=index,
            token=token,
            classification=classification,
            visible=visible,
            confidence=confidence,
            user_attrs=user_attrs,
        )

    def add_inference_2d_bbox(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        top: Union[int, float],
        left: Union[int, float],
        width: Union[int, float],
        height: Union[int, float],
        confidence: float,
        area: Optional[float] = None,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add an inference for a 2D bounding box.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            top (int or float): The top of the box in pixels
            left (int or float): The left of the box in pixels
            width (int or float): The width of the box in pixels
            height (int or float): The height of the box in pixels
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            area (float, optional): The area of the image.
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """

        self.add_2d_bbox(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            top=top,
            left=left,
            width=width,
            height=height,
            confidence=confidence,
            area=area,
            iscrowd=iscrowd,
            user_attrs=user_attrs,
        )

    def add_inference_2d_line_segment(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        x1: Union[int, float],
        y1: Union[int, float],
        x2: Union[int, float],
        y2: Union[int, float],
        confidence: float,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add an inference for a 2D line segment.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            x1 (int or float): The x-coord of the first vertex in pixels
            y1 (int or float): The x-coord of the first vertex in pixels
            x2 (int or float): The x-coord of the first vertex in pixels
            y2 (int or float): The x-coord of the first vertex in pixels
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_2d_line_segment(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            x1=x1,
            y1=y1,
            x2=x2,
            y2=y2,
            confidence=confidence,
            iscrowd=iscrowd,
            user_attrs=user_attrs,
            links=links,
        )

    def add_inference_3d_cuboid(
        self,
        *,
        label_id: str,
        classification: str,
        position: List[float],
        dimensions: List[float],
        rotation: List[float],
        confidence: float,
        iscrowd: Optional[bool] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
        links: Optional[Dict[str, Any]] = None,
        coord_frame_id: Optional[str] = None,
    ) -> None:
        """Add an inference for a 3D cuboid.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            position (list of float): the position of the center of the cuboid
            dimensions (list of float): the dimensions of the cuboid
            rotation (list of float): the local rotation of the cuboid, represented as an xyzw quaternion.
            confidence (float): confidence of prediction
            iscrowd (bool, optional): Is this label marked as a crowd. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
            links (dict, optional): Links between labels. Defaults to None.
            coord_frame_id (str, optional): Coordinate frame id. Defaults to 'world'
        """
        self.add_3d_cuboid(
            label_id=label_id,
            classification=classification,
            position=position,
            dimensions=dimensions,
            rotation=rotation,
            confidence=confidence,
            iscrowd=iscrowd,
            user_attrs=user_attrs,
            links=links,
            coord_frame_id=coord_frame_id,
        )

    def add_inference_2d_keypoints(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        keypoints: List[Dict[KEYPOINT_KEYS, Union[int, float, str]]],
        confidence: float,
        top: Optional[Union[int, float]] = None,
        left: Optional[Union[int, float]] = None,
        width: Optional[Union[int, float]] = None,
        height: Optional[Union[int, float]] = None,
        polygons: Optional[
            List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]]
        ] = None,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add an inference for a 2D keypoints task. Can include a bounding box and/or polygon.

        A keypoint is a dictionary of the form:
            'x': x-coordinate in pixels
            'y': y-coordinate in pixels
            'name': string name of the keypoint

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            keypoints (list of dicts): The keypoints of this detection
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            top (int or float, optional): The top of the bounding box in pixels. Defaults to None.
            left (int or float, optional): The left of the bounding box in pixels. Defaults to None.
            width (int or float, optional): The width of the bounding box in pixels. Defaults to None.
            height (int or float, optional): The height of the bounding box in pixels. Defaults to None.
            polygons (list of dicts, optional): The polygon geometry. Defaults to None.
            center (list of ints or floats, optional): The center point of the polygon instance. Defaults to None.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_2d_keypoints(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            keypoints=keypoints,
            confidence=confidence,
            top=top,
            left=left,
            width=width,
            height=height,
            polygons=polygons,
            center=center,
            user_attrs=user_attrs,
        )

    def add_inference_2d_polygon_list(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        polygons: List[Dict[POLYGON_VERTICES_KEYS, List[Tuple[Union[int, float]]]]],
        confidence: float,
        center: Optional[List[Union[int, float]]] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add an inference for a 2D polygon list instance segmentation task.

        Polygons are dictionaries of the form:
            'vertices': List of (x, y) vertices (e.g. [[x1,y1], [x2,y2], ...])
                The polygon does not need to be closed with (x1, y1).
                As an example, a bounding box in polygon representation would look like:

                .. code-block::

                    {
                        'vertices': [
                            [left, top],
                            [left + width, top],
                            [left + width, top + height],
                            [left, top + height]
                        ]
                    }


        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            polygons (list of dicts): The polygon geometry
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            center (list of ints or floats, optional): The center point of the instance
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_2d_polygon_list(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            polygons=polygons,
            confidence=confidence,
            center=center,
            user_attrs=user_attrs,
        )

    def add_inference_2d_instance_seg(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        instance_mapping: List[Dict[str, Any]],
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """Add an inference for 2D instance segmentation.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            instance_mapping (List[Dict]): a list of instances present in the mask,
                each is a numeric `id`, string `classification`, and optional dict of additional attributes.
                As an example of one instance:
                .. code-block::
                        {
                            'id': 1,
                            'classification': "Person",
                            'attributes': {
                                'is_standing': false,
                            }
                        }
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """
        self.add_2d_instance_seg(
            sensor_id=sensor_id,
            label_id=label_id,
            mask_url=mask_url,
            mask_data=mask_data,
            instance_mapping=instance_mapping,
            resize_mode=resize_mode,
        )

    def add_inference_2d_semseg(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        mask_url: Optional[str] = None,
        mask_data: Optional[Any] = None,
        resize_mode: Optional[ResizeMode] = None,
    ) -> None:
        """Add an inference for 2D semseg.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            mask_url (str, optional): URL to the pixel mask png.
            mask_data (ndarray, optional): ndarray of pixel mask data, shape [height, width]
            resize_mode (ResizeMode, optional): if the mask is a different size from the base image, define how to display it
                "fill": stretch the mask to fit the base image dimensions
                None: no change
        """
        self.add_2d_semseg(
            sensor_id=sensor_id,
            label_id=label_id,
            mask_url=mask_url,
            mask_data=mask_data,
            resize_mode=resize_mode,
        )

    def add_inference_2d_classification(
        self,
        *,
        sensor_id: str = DEFAULT_SENSOR_ID,
        label_id: str,
        classification: str,
        confidence: float,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add an inference for 2D classification.

        Args:
            sensor_id (str): ID of sensor that generates this datapoint. Must match base frame sensor_id. Do not use arg if not used in base frame.
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_2d_classification(
            sensor_id=sensor_id,
            label_id=label_id,
            classification=classification,
            confidence=confidence,
            user_attrs=user_attrs,
        )

    def add_inference_3d_classification(
        self,
        *,
        label_id: str,
        classification: str,
        confidence: float,
        coord_frame_id: Optional[str] = None,
        user_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Add a label for 3D classification.

        Args:
            label_id (str): label_id which is unique across datasets and inferences.
            classification (str): the classification string
            confidence (float): The confidence between 0.0 and 1.0 of the prediction
            coord_frame_id (optional, str): The coordinate frame id.
            user_attrs (dict, optional): Any additional label-level metadata fields. Defaults to None.
        """
        self.add_3d_classification(
            label_id=label_id,
            classification=classification,
            confidence=confidence,
            coord_frame_id=coord_frame_id,
            user_attrs=user_attrs,
        )


class Inferences:
    """A container used to construct a set of inferences.

    Typical usage is to create an Inferences object, add multiple InferencesFrames to it,
    then serialize the frames to be submitted.
    """

    _frames: List[InferencesFrame]
    _frame_ids_set: Set[str]
    _label_ids_set: Set[str]
    _label_classes_set: Set[str]
    _label_class_ids_set: Set[int]
    _frame_summaries: List[InferenceFrameSummary]
    _temp_frame_file_names: List[str]
    _temp_frame_asset_file_names: List[str]
    _temp_frame_embeddings_file_names: List[str]
    backfilled_frame_embs_flag: bool
    backfilled_crops_flag: bool
    crop_embedding_dim: Optional[int]
    frame_embedding_dim: Optional[int]

    def __init__(self) -> None:
        if not _is_one_gb_available():
            raise OSError(
                "Attempting to run with less than 1 GB of available disk space. Exiting..."
            )
        self._frames = []
        self._frame_ids_set = set()
        self._label_ids_set = set()
        self._label_classes_set = set()
        self._label_class_ids_set = set()
        self._frame_summaries = []
        current_time = datetime.datetime.now()
        self.temp_file_path = create_temp_directory()
        self._temp_frame_prefix = "al_{}_inference_".format(
            current_time.strftime("%Y%m%d_%H%M%S_%f")
        )
        self._temp_frame_assets_prefix = "al_{}_dataset_".format(
            current_time.strftime("%Y%m%d_%H%M%S_%f")
        )
        self._temp_frame_embeddings_prefix = "al_{}_inference_embeddings_".format(
            current_time.strftime("%Y%m%d_%H%M%S_%f")
        )
        self._temp_frame_file_names = []
        self._temp_frame_embeddings_file_names = []
        self._temp_frame_asset_file_names = []
        self.backfilled_frame_embs_flag = False
        self.backfilled_crops_flag = False
        self.frame_embedding_dim = None
        self.crop_embedding_dim = None

    def _cleanup_temp_dir(self) -> None:
        mark_temp_directory_complete(self.temp_file_path)

    def _save_rows_to_temp(
        self,
        file_name_prefix: str,
        writefunc: Callable[[IO[Any]], None],
        mode: str = "w",
    ) -> Optional[str]:
        """[summary]

        Args:
            file_name_prefix (str): prefix for the filename being saved
            writefunc ([filelike): function used to write data to the file opened

        Returns:
            str or None: path of file or none if nothing written
        """

        if not _is_one_gb_available():
            raise OSError(
                "Attempting to flush inferences to disk with less than 1 GB of available disk space. Exiting..."
            )

        data_rows_content = NamedTemporaryFile(
            mode=mode, delete=False, prefix=file_name_prefix, dir=self.temp_file_path
        )
        data_rows_content_path = data_rows_content.name
        writefunc(data_rows_content)

        # Nothing was written, return None
        if data_rows_content.tell() == 0:
            return None

        data_rows_content.seek(0)
        data_rows_content.close()
        return data_rows_content_path

    def _backfill_frame_embeddings_if_necessary(self) -> bool:
        """Determines if there are provided crop embeddings but no frame embeddings
        Backfills frame embeddings with dummy [0.0, 0.0, 0.0] if crop embeddings exist
        Returns True if backfill was necessary

        Returns:
            bool: returns if backfill on frame was necessary
        """
        if all([len(inf_frame.crop_embeddings) == 0 for inf_frame in self._frames]):
            return False

        count = len([frame for frame in self._frames if frame.embedding is not None])
        if count != 0:
            return False

        for frame in self._frames:
            frame.add_frame_embedding(embedding=DUMMY_FRAME_EMBEDDING)

        return True

    def _flush_to_disk(self) -> None:
        """Writes the all the frames in the frame buffer to temp file on disk"""
        self.backfilled_frame_embs_flag = self._backfill_frame_embeddings_if_necessary()

        if len(self._frames) == 0:
            return
        frame_path = self._save_rows_to_temp(
            self._temp_frame_prefix, lambda x: self.write_to_file(x)
        )
        if frame_path:
            self._temp_frame_file_names.append(frame_path)
        embeddings_path = self._save_rows_to_temp(
            self._temp_frame_embeddings_prefix,
            lambda x: self.write_embeddings_to_file(x),
            mode="wb",
        )
        if embeddings_path:
            self._temp_frame_embeddings_file_names.append(embeddings_path)

        assets_path = self._save_rows_to_temp(
            self._temp_frame_assets_prefix,
            lambda x: self.write_label_assets_to_file(x),
            mode="wb",
        )
        if assets_path:
            self._temp_frame_asset_file_names.append(assets_path)

        self._frames = []

    def _validate_frame(
        self, frame_summary: InferenceFrameSummary, project_info: Dict[str, Any]
    ) -> None:
        """Validates single frame in set according to project constraints

        Args:
            frame_summary Dict[str, Any]: dictionary representation of an InferencesFrame summary
            project_info Dict[str, Any]: metadata about the project being uploaded to
        """

        frame_id = frame_summary["frame_id"]
        primary_task = project_info.get("primary_task")

        # TODO: Should also check this for classification here.
        # 2D_SEMSEG frames may only have one label of type SEMANTIC_LABEL_URL_2D
        if primary_task == "2D_SEMSEG":
            label_counts = frame_summary["label_counts"]
            url_count = label_counts.get("SEMANTIC_LABEL_URL_2D", 0)
            assets_count = label_counts.get("SEMANTIC_LABEL_ASSET_2D", 0)
            if url_count + assets_count != 1:
                extra_labels = filter(
                    lambda x: x != "SEMANTIC_LABEL_URL_2D"
                    and x != "SEMANTIC_LABEL_ASSET_2D",
                    label_counts.keys(),
                )
                issue_text = (
                    "has no labels"
                    if not len(label_counts)
                    else f"has label types of {list(extra_labels)}"
                )
                raise Exception(
                    f"Frame {frame_id} {issue_text}. Inferences frames for 2D_SEMSEG projects must have exactly one 2d_semseg inference "
                )

        if primary_task == "2D_INSTANCE_SEGMENTATION":
            label_counts = frame_summary["label_counts"]
            url_count = label_counts.get("INSTANCE_LABEL_URL_2D", 0)
            assets_count = label_counts.get("INSTANCE_LABEL_ASSET_2D", 0)
            if url_count + assets_count != 1:
                extra_labels = filter(
                    lambda x: x != "INSTANCE_LABEL_URL_2D"
                    and x != "INSTANCE_LABEL_ASSET_2D",
                    label_counts.keys(),
                )
                issue_text = (
                    "has no labels"
                    if not len(label_counts)
                    else f"has label types of {list(extra_labels)}"
                )
                raise Exception(
                    f"Frame {frame_id} {issue_text}. Inferences frames for 2D_INSTANCE_SEGMENTATION projects must have exactly one 2d_instance_seg inference"
                )

        project_custom_metrics_names = [
            custom_metric["name"]
            for custom_metric in project_info.get("custom_metrics", []) or []
        ]

        for inference_custom_metric_name in frame_summary["custom_metrics_names"]:
            if inference_custom_metric_name not in project_custom_metrics_names:
                potential_candidates_text = ""

                if not len(project_custom_metrics_names):
                    potential_candidates_text = (
                        "No custom metrics have been configured for this project"
                    )

                elif len(project_custom_metrics_names) < 11:
                    potential_candidates_text = (
                        "This project does have the following custom metrics: "
                        f"{', '.join(project_custom_metrics_names)}"
                    )
                else:
                    potential_candidates_text = (
                        "This project includes the following custom metrics: "
                        f"{', '.join(project_custom_metrics_names[:10])}. A full list is returned from calling get_project."
                    )

                raise Exception(
                    f"Frame {frame_id} contains a custom metric {inference_custom_metric_name} that is not configured "
                    f"for project {project_info['id']}. {potential_candidates_text}"
                )

    def _validate_frames(self, project_info: Dict[str, Any]) -> None:
        """Validates all frames in set according to project constraints

        Args:
            project_info Dict[str, Any]: metadata about the project being uploaded to
        """
        for frame_summary in self._frame_summaries:
            self._validate_frame(frame_summary, project_info)

        primary_task = project_info.get("primary_task")
        if primary_task == "BINARY_CLASSIFICATION":
            negative_class = project_info.get("binary_classification_negative_class")
            classifications = self._label_classes_set
            if negative_class in classifications:
                raise Exception(
                    f"Inferences must be submitted as the positive class with confidence 0 to 1 for binary classification project {project_info['id']}. Inferences will be interpreted by Aquarium using the decision threshold."
                )

        if self.backfilled_crops_flag:
            print(
                "WARNING: One or more Frames contained Frame Embeddings with Crops with no Crop Embeddings. All crops without Crop Embeddings have been backfilled with a default Embedding Vector"
            )
        if self.backfilled_frame_embs_flag:
            print(
                "WARNING: Crop Embeddings were provided but no Frame Embeddings were provided. All inf frames have backfilled with a default Frame Embedding Vector"
            )

    def add_frame(self, frame: InferencesFrame) -> None:
        """Add an InferencesFrame to this dataset.

        Args:
            frame (InferencesFrame): An InferencesFrame in this dataset.
        """
        if not isinstance(frame, InferencesFrame):
            raise Exception("Frame is not an InferencesFrame")

        if frame.frame_id in self._frame_ids_set:
            raise Exception("Attempted to add duplicate frame id.")

        duplicate_label_ids = frame._label_ids_set & self._label_ids_set
        if duplicate_label_ids:
            raise Exception(
                f"Attempted to add duplicate label id(s): {duplicate_label_ids}"
            )

        self.backfilled_crops_flag = frame.validate_and_backfill_crop_embeddings()
        self._frames.append(frame)
        self._frame_ids_set.add(frame.frame_id)
        self._label_ids_set.update(frame._label_ids_set)
        self._label_class_ids_set.update(frame._seen_label_class_ids)
        self._label_classes_set.update(frame._all_label_classes())
        self._frame_summaries.append(frame._to_summary())

        if self.frame_embedding_dim is None:
            self.frame_embedding_dim = frame.frame_embedding_dim
        elif self.frame_embedding_dim != frame.frame_embedding_dim:
            raise Exception(
                f"Length of frame embeddings must be the same, existing embeddings are of dimension {self.frame_embedding_dim} but new embedding has dimension {frame.frame_embedding_dim}"
            )

        if len(frame._label_ids_set) > 0:
            if len(frame.crop_embeddings) > 0 and len(frame.crop_embeddings) != len(
                frame._label_ids_set
            ):
                raise Exception(
                    f"Embeddings must be provided for all crops or no crops. Embeddings are provided for partial crops for frame {frame.frame_id}."
                )

            crop_embedding_dim = (
                frame.crop_embedding_dim if frame.crop_embedding_dim else -1
            )
            if self.crop_embedding_dim is None:
                self.crop_embedding_dim = crop_embedding_dim
            elif self.crop_embedding_dim != crop_embedding_dim:
                raise Exception(
                    f"Length of crop embeddings must be the same, existing embeddings are of dimension {self.crop_embedding_dim} but new embedding has dimension {crop_embedding_dim}"
                )

        if len(self._frames) >= MAX_FRAMES_PER_BATCH:
            self._flush_to_disk()

    def delete_frame(self, frame_id: str) -> None:
        """Delete a frame from this dataset.

        Args:
            frame_id (str): id of frame to delete.
        """

        if frame_id in self._frame_ids_set:
            raise Exception("Attempted to add and delete same frame id.")

        self._frames.append(
            InferencesFrame(
                frame_id=frame_id, reuse_latest_embedding=True, update_type="DELETE"
            )
        )
        self._frame_ids_set.add(frame_id)

        if len(self._frames) >= MAX_FRAMES_PER_BATCH:
            self._flush_to_disk()

    def write_to_file(self, filelike: IO[Any]) -> None:
        """Write the frame content to a text filelike object (File handle, StringIO, etc.)

        Args:
            filelike (filelike): The destination file-like to write to.
        """

        for frame in self._frames:
            row = frame.to_dict()
            filelike.write(json.dumps(row) + "\n")

    def write_label_assets_to_file(self, filelike: IO[Any]) -> None:
        """Write the frame's label assets to a text filelike object (File handle, StringIO, etc.)

        Args:
            filelike (filelike): The destination file-like to write to.
        """
        count = len(self._frames)
        with_label_assets_count = len(
            [frame for frame in self._frames if len(frame.label_assets) > 0]
        )
        if with_label_assets_count == 0:
            return

        frame_ids = np.empty((count), dtype=object)
        label_assets = np.empty((count), dtype=object)

        for i, frame in enumerate(self._frames):
            frame_ids[i] = frame.frame_id
            label_assets[i] = frame.label_assets

        df = pd.DataFrame(
            {"frame_ids": pd.Series(frame_ids), "label_assets": pd.Series(label_assets)}
        )

        arrow_data = pa.Table.from_pandas(df)
        writer = pa.ipc.new_file(filelike, arrow_data.schema, use_legacy_format=False)
        writer.write(arrow_data)
        writer.close()

    def write_embeddings_to_file(self, filelike: IO[Any]) -> None:
        """Write the frame's embeddings to a text filelike object (File handle, StringIO, etc.)

        Args:
            filelike (filelike): The destination file-like to write to.
        """

        count = len([frame for frame in self._frames if frame.embedding is not None])

        if count == 0 or not self._frames:
            return

        if count != len(self._frames):
            raise Exception(
                "If any frames have user provided embeddings, all frames must have embeddings."
            )

        # TODO: Is there a pattern like type predicate functions we can use
        # to avoid error-prone inline manual casts like this?
        frames_with_embs = cast(List[InferenceFrameWithEmbeddings], self._frames)

        # Get the first frame embedding dimension
        frame_embedding_dim = len(frames_with_embs[0].embedding["embedding"])
        # Get the first crop embedding dimension
        crop_embedding_dim = 1
        for frame in frames_with_embs:
            if frame.crop_embeddings:
                first_crop_emb = frame.crop_embeddings[0]
                crop_embedding_dim = len(first_crop_emb["embedding"])
                break

        frame_ids = np.empty((count), dtype=object)
        frame_embeddings = np.empty((count), dtype=object)
        crop_ids = np.empty((count), dtype=object)
        crop_embeddings = np.empty((count), dtype=object)

        for i, frame in enumerate(frames_with_embs):
            frame_ids[i] = frame.embedding["task_id"]
            frame_embeddings[i] = frame.embedding["embedding"]
            crop_ids[i] = [x["uuid"] for x in frame.crop_embeddings]
            crop_embeddings[i] = [x["embedding"] for x in frame.crop_embeddings]

        df = pd.DataFrame(
            {
                "frame_ids": pd.Series(frame_ids),
                "frame_embeddings": pd.Series(frame_embeddings),
                "crop_ids": pd.Series(crop_ids),
                "crop_embeddings": pd.Series(crop_embeddings),
            }
        )

        arrow_data = pa.Table.from_pandas(df)
        writer = pa.ipc.new_file(filelike, arrow_data.schema, use_legacy_format=False)
        writer.write(arrow_data)
        writer.close()
